/*
 * testDivMprec_g.cu
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2013 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * test file for division of doubles on the GPU
 */

#include <ctime>
#include <cstdio>
#include <gmp.h> 
#include <iostream>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpfr.h>
#include <mpfi.h>
#include "mpfi_io.h"
#include "../../src/multi_prec.h"

#define K 5
#define L 4
#define R 4

//computes log2 from x
inline long double log2(const long double x){
    return  log(x) * M_LOG2E;
}

__global__ void test(double* buff1, double* buff2, double* res){
	/********* MPrec computations *********/
  multi_prec<K, double> mp1;
  multi_prec<L, double> mp2;
  multi_prec<R, double> mpRes;

  mp1.setData(buff1, K);
  mp2.setData(buff2, L);

	divNewton(mpRes, mp1, mp2);

	for(int i=0; i<R; i++)
		res[i] = mpRes.getData()[i];
}

int check(int prec, int j, double* buff1, double* buff2, double* mpRes){
	int retVal = 0;
	int q = floor(log2((double)R));
	mpfr_t mpf1, mpf2, mpfRes, mpfMP, mpfErr, mpfBound;
  mpfr_init2(mpf1, prec*53 + 106);
  mpfr_init2(mpf2, prec*53 + 106);
  mpfr_init2(mpfMP, prec*53 + 106);
  mpfr_init2(mpfRes, prec*53 + 106);
  mpfr_init2(mpfErr, prec*53 + 106);
	mpfr_init2(mpfBound, prec*53 + 106);

	mpfr_set_d(mpfBound, q, GMP_RNDN);
  mpfr_exp2(mpfBound, mpfBound, GMP_RNDN);
  mpfr_mul_ui(mpfBound, mpfBound, 50, GMP_RNDN);
  mpfr_neg(mpfBound, mpfBound, GMP_RNDN);
  mpfr_sub_ui(mpfBound, mpfBound, 1, GMP_RNDN);
  mpfr_exp2(mpfBound, mpfBound, GMP_RNDN);

  /********* MPFR computations *********/
  mpfr_set_d(mpf1, buff1[0], GMP_RNDN);
  for(int i=1; i<K; i++)
    mpfr_add_d(mpf1, mpf1, buff1[i], GMP_RNDN);
  mpfr_set_d(mpf2, buff2[0], GMP_RNDN);
  for(int i=1; i<L; i++)
    mpfr_add_d(mpf2, mpf2, buff2[i], GMP_RNDN);

  mpfr_div(mpfRes, mpf1, mpf2, GMP_RNDN);

  /********* Error computation *********/
	mpfr_set_d(mpfMP, mpRes[0] , GMP_RNDN);
  for(int i=1; i<R; i++)
    mpfr_add_d(mpfMP, mpfMP, mpRes[i], GMP_RNDN);

  mpfr_sub(mpfErr, mpfMP, mpfRes, GMP_RNDN);
  mpfr_div(mpfErr, mpfErr, mpfRes, GMP_RNDN);
  mpfr_abs(mpfErr, mpfErr, GMP_RNDN);

  if( (mpfr_cmp(mpfErr, mpfBound) <= 0) ){}
  else{
    cout << "------- Test " << j << " ---------" << endl;

    cout << "------- MPrec Input ---------" << endl;
    printArrayBin<K>(buff1);
    printArrayBin<L>(buff2);

    cout << "------- MPrec Output ---------" << endl;
    printArrayBin<R>(mpRes);

    cout << "------- MPFR Input ---------" << endl;
    mpfr_out_str(stdout, 2, 0, mpf1, GMP_RNDN);
    cout << endl;
    mpfr_out_str(stdout, 2, 0, mpf2, GMP_RNDN);
    cout << endl;

    cout << "------- MPFR Output ---------" << endl;
    mpfr_out_str(stdout, 2, 0, mpfRes, GMP_RNDN);
    cout << endl;

		// \abs{ \frac{x-a^{-1}}{a^{-1}} }
    cout << "------- MPFR Error ---------" << endl;
    mpfr_out_str(stdout, 2, 0, mpfErr, GMP_RNDN);
    cout << endl;

    // 2^{-2^q(p-3)-1}
    cout << "------- MPFR Bound ---------" << endl;
    mpfr_out_str(stdout, 2, 0, mpfBound, GMP_RNDN);
    cout << endl;

    retVal = 1;
  }

	mpfr_clear(mpf1);
  mpfr_clear(mpf2);
  mpfr_clear(mpfMP);
  mpfr_clear(mpfRes);
  mpfr_clear(mpfErr);
	mpfr_clear(mpfBound);
	return retVal;
}

int main(int argc, char *argv[]) {
	double buff1[K], buff2[L], res[R];
  int prec = (K>L)?K:L;
  prec = (R>prec)?R:prec;
	int failCases = 0;

	if(argc<2){
    cout << "Usage: ./testDivMprec_g trials" << endl;
    return 0;
  }

  int trials = atoi(argv[1]);

	cout << "Testing " << trials << " MPrec divisions with: " << K << " doubles / " << L << " doubles = " << R << " doubles " << endl;

	double *dev_buff1, *dev_buff2, *dev_res;
  cudaMalloc((void**)&dev_buff1, K * sizeof(double));
  cudaMalloc((void**)&dev_buff2, L * sizeof(double));
  cudaMalloc((void**)&dev_res, R * sizeof(double));

	for (int i=0; i<trials; i++){
		generateNormalizedExpansion<K,double>(buff1);
		generateNormalizedExpansion<L,double>(buff2);

		cudaMemcpy(dev_buff1,buff1, K * sizeof(double),cudaMemcpyHostToDevice);
		cudaMemcpy(dev_buff2,buff2, L * sizeof(double),cudaMemcpyHostToDevice);

		test<<<1,1>>>(dev_buff1, dev_buff2, dev_res);

		cudaMemcpy(res,dev_res, R * sizeof(double),cudaMemcpyDeviceToHost);
		failCases += check(prec, i, buff1, buff2, res);
	}

	cout << "Total failed cases:" << failCases << endl;

	cudaFree(dev_buff1);
	cudaFree(dev_buff2);
	cudaFree(dev_res);

  return 0;
}

