/*
 * testDivMprec_c.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/*
 * test file for division of floats on the CPU
 */

#include <ctime>
#include <cstdio>
#include <iostream>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <mpfr.h>
#include <mpfi.h>
#include "mpfi_io.h"
#include "../../src/multi_prec.h"

#define K 4
#define L 4
#define R 4

//computes log2 from x
inline long double log2(const long double x){
    return  log(x) * M_LOG2E;
}

int test(int trials){
	float buff1[K], buff2[L];
	int failCases = 0;
	int prec = (K>L)?K:L;
	prec = (R>prec)?R:prec;
	int q = floor(log2((double)R));

	mpfr_t mpf1, mpf2, mpfRes, mpfMP, mpfErr, mpfBound;
	mpfr_init2(mpf1, prec*24 + 48);
	mpfr_init2(mpf2, prec*24 + 48);
	mpfr_init2(mpfMP, prec*24 + 48);
	mpfr_init2(mpfRes, prec*24 + 48);
	mpfr_init2(mpfErr, prec*24 + 48);
	mpfr_init2(mpfBound, prec*24 + 48);

  mpfr_set_d(mpfBound, q, GMP_RNDN);
  mpfr_exp2(mpfBound, mpfBound, GMP_RNDN);
  mpfr_mul_ui(mpfBound, mpfBound, 21, GMP_RNDN);
  mpfr_neg(mpfBound, mpfBound, GMP_RNDN);
  mpfr_sub_ui(mpfBound, mpfBound, 1, GMP_RNDN);
  mpfr_exp2(mpfBound, mpfBound, GMP_RNDN);

	multi_prec<K, float> mp1;
	multi_prec<L, float> mp2;
	multi_prec<R, float> mpRes;

	for (int j=1; j<=trials; j++){
	  generateNormalizedExpansion<K>(buff1);
		generateNormalizedExpansion<L>(buff2);

		/********* MPrec computations *********/
		mp1.setData(buff1, K);
    mp2.setData(buff2, L);

		divNewton(mpRes, mp1, mp2);

		/********* MPFR computations *********/
	  mpfr_set_d(mpf1, buff1[0], GMP_RNDN);
	  for(int i=1; i<K; i++)
 	  	mpfr_add_d(mpf1, mpf1, buff1[i], GMP_RNDN);
	  mpfr_set_d(mpf2, buff2[0], GMP_RNDN);
	  for(int i=1; i<L; i++)
	  	mpfr_add_d(mpf2, mpf2, buff2[i], GMP_RNDN);

	  mpfr_div(mpfRes, mpf1, mpf2, GMP_RNDN);

		/********* Error computation *********/
	  mpfr_set_d(mpfMP, mpRes[0] , GMP_RNDN);
	  for(int i=1; i<R; i++)
	  	mpfr_add_d(mpfMP, mpfMP, mpRes[i], GMP_RNDN);

		mpfr_sub(mpfErr, mpfMP, mpfRes, GMP_RNDN);
    mpfr_div(mpfErr, mpfErr, mpfRes, GMP_RNDN);
    mpfr_abs(mpfErr, mpfErr, GMP_RNDN);

	  if( (mpfr_cmp(mpfErr, mpfBound) <= 0) ){}
	  else{
			cout << "------- Test " << j << " ---------" << endl;

			cout << "------- MPrec Input ---------" << endl;
    	mp1.printMultiPrecBin();
    	mp2.printMultiPrecBin();

			cout << "------- MPrec Output ---------" << endl;
    	mpRes.printMultiPrecBin();

			cout << "------- MPFR Input ---------" << endl;
    	mpfr_out_str(stdout, 2, 0, mpf1, GMP_RNDN);
    	cout << endl;
    	mpfr_out_str(stdout, 2, 0, mpf2, GMP_RNDN);
    	cout << endl;

			cout << "------- MPFR Output ---------" << endl;
    	mpfr_out_str(stdout, 2, 0, mpfRes, GMP_RNDN);	    
			cout << endl;

			// \abs{ \frac{x-a^{-1}}{a^{-1}} }
      cout << "------- MPFR Error ---------" << endl;
      mpfr_out_str(stdout, 2, 0, mpfErr, GMP_RNDN);
      cout << endl;

      // 2^{-2^q(p-3)-1}
      cout << "------- MPFR Bound ---------" << endl;
      mpfr_out_str(stdout, 2, 0, mpfBound, GMP_RNDN);
      cout << endl;

			failCases++;
	  }
	}

	mpfr_clear(mpf1);
	mpfr_clear(mpf2);
	mpfr_clear(mpfMP);
	mpfr_clear(mpfRes);
	mpfr_clear(mpfErr);
  mpfr_clear(mpfBound);
  return failCases;
}

int main(int argc, char *argv[]) {
	if(argc<2){
		cout << "Usage: ./testDivMprec_c trials" << endl; 
   	return 0;
  }

	int trials = atoi(argv[1]);

  cout << "Testing " << trials << " MPrec divisions with: " << K << " floats / " << L << " floats = " << R << " floats " << endl;
  int failCases = test(trials);
  cout << "Total failed cases:" << failCases << endl;
  return 0;
}
