/*
 * timingInv_c.cu - Bailey style testing for reciprocal
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2013 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu popescu@laas.fr
 *               Mioara Joldes joldes@laas.fr
 */
/* Test file for reciprocal of mprec using doubles*/

#include <ctime>
#include <cstdio>
#include <gmp.h>
#include <iostream>
#include <iomanip>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpfr.h>
#include <mpfi.h>
#include "mpfi_io.h"
#include "../src/multi_prec.h"
#define prec 16 //prec down = denominator
#define precu 8 //prec output 

/*We test the performance of multi_prec reciprocal by taking random generated nonoverlapping expansions and then inverting them;
The point is to get performance measurements in MFlops for the following combinations:
-- 1/d
-- 1/2d
-- 1/4d
-- etc;
*/

const double MICRO(1e-6);       // Used for the timing.
//----------------------------------------------------------------------
// A basic timer that works for OMP code.
//----------------------------------------------------------------------
double my_timer(timeval &tim, double &t_start, double &t_stop,
                          const string &command){
	if ( 0 == command.compare("start") ) {
  	gettimeofday(&tim, NULL);
    t_start = tim.tv_sec+(tim.tv_usec*MICRO);
  }
  if ( 0 == command.compare("stop") ) {
    gettimeofday(&tim, NULL);
    t_stop = tim.tv_sec+(tim.tv_usec*MICRO);
    cout << "Wall time: " << setprecision(8) << fixed
         << t_stop - t_start << " seconds.";
  }
  return (t_stop - t_start);
}

double testInv(int trials){
	multi_prec<prec, double> mpa, ma1, ma2, ma3, ma4;
	multi_prec<precu,double> b1, b2, b3, b4;
	double buff[prec];
  generateNormalizedExpansion<prec>(buff);
  mpa.setData(buff,prec);
//	mpa.printMultiPrecBin();
	ma1 = 1.0 + (1.0/trials) * mpa;
	ma2 = 1.0 + (2.0/trials) * mpa;
	ma3 = 1.0 + (3.0/trials) * mpa;
  ma4 = 1.0 + (4.0/trials) * mpa;	
	b1 = 1.0;
  b2 = 1.0;
	b3 = 1.0;
	b4 = 1.0;

  // Capture the start time.
  timeval tim;
  double t_start, t_stop;
  my_timer(tim, t_start, t_stop, "start");
       
	for (int i=0; i<trials; i++){
		invExpansion(b1, ma1);
		invExpansion(b2, ma2);
		invExpansion(b3, ma3);
		invExpansion(b4, ma4);
	}
	double seconds = my_timer(tim, t_start, t_stop, "stop");
  cout << " Performance: " << 4*1e-6*trials/seconds << " Mops/second." << endl;

  //end timing      
	b1 = (b1+b2+b3+b4);
  return b1[0];
}

int main(int argc, char *argv[]) {
	if(argc<2){
		cout << "Usage: ./timingInv trials" << endl; 
   	return 0;
 	}

	int trials = atoi(argv[1]);

  cout << "Testing " << trials << " inversions 1.0/" << prec << "d and outputting " << precu << "d precision." << endl;
  double res = testInv(trials);
  cout << "Result: " << res << endl;
  return 0;
}
