/*
 * timingAdd_c.cu - Bailey style testing for addition
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2013 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@univ-tlse3.fr
 *               Mioara Joldes joldes@laas.fr
 */
/* Test file for addition of mprec using doubles */

#include <ctime>
#include <cstdio>
#include <gmp.h>
#include <iostream>
#include <iomanip>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpfr.h>
#include <mpfi.h>
#include "mpfi_io.h"
#include "../src/multi_prec.h"
#define prec 4 //prec 
#define precu 16 //prec result

/*We test the performance of multi_prec addition by taking random generated nonoverlapping expansions and then adding them;
The point is to get performance measurements in MFlops for the following combinations:
-- 1d/2d
-- 2d/2d
-- 4d/2d
-- etc;

-- 1d/4d
-- 2d/4d
-- 4d/4d
-- etc.
*/

const double MICRO(1e-6);// Used for the timing.
//----------------------------------------------------------------------
// A basic timer that works for OMP code.
//----------------------------------------------------------------------
double my_timer(timeval &tim, double &t_start, double &t_stop,
                          const string &command){
  if ( 0 == command.compare("start") ) {
  	gettimeofday(&tim, NULL);
    t_start = tim.tv_sec+(tim.tv_usec*MICRO);
  }
  if ( 0 == command.compare("stop") ) {
    gettimeofday(&tim, NULL);
    t_stop = tim.tv_sec+(tim.tv_usec*MICRO);
    cout << "Wall time: " << setprecision(8) << fixed
         << t_stop - t_start << " seconds.";
  }
  return (t_stop - t_start);
}

double testAdd(int trials){
	multi_prec<prec, double> mpa, ma1, ma2, ma3, ma4;
	multi_prec<precu,double> b1, b2, b3, b4;
	double buff[prec];
	generateNormalizedExpansion<prec>(buff);
  mpa.setData(buff,prec);
//	mpa.printMultiPrecBin();
	ma1 = 1.0 + (1.0/trials) * mpa;
	ma2 = 1.0 + (2.0/trials) * mpa;
	ma3 = 1.0 + (3.0/trials) * mpa;
  ma4 = 1.0 + (4.0/trials) * mpa;	
	b1 = 1.0;
  b2 = 1.0;
	b3 = 1.0;
	b4 = 1.0;

	mpfr_t mpfa1, mpfa2, mpfa3, mpfa4;
	mpfr_init2(mpfa1, prec*53);
  mpfr_init2(mpfa2, prec*53);
  mpfr_init2(mpfa3, prec*53);
  mpfr_init2(mpfa4, prec*53);

	mpfr_set_d(mpfa1, ma1[0], GMP_RNDN);
	mpfr_set_d(mpfa2, ma2[0], GMP_RNDN);
	mpfr_set_d(mpfa3, ma3[0], GMP_RNDN);
	mpfr_set_d(mpfa4, ma4[0], GMP_RNDN);
  for(int i=1; i<prec; i++){
  	mpfr_add_d(mpfa1, mpfa1, ma1[i], GMP_RNDN);
		mpfr_add_d(mpfa2, mpfa2, ma2[i], GMP_RNDN);
		mpfr_add_d(mpfa3, mpfa3, ma3[i], GMP_RNDN);
		mpfr_add_d(mpfa4, mpfa4, ma4[i], GMP_RNDN);
	}

	mpfr_t mpfb1, mpfb2, mpfb3, mpfb4;
  mpfr_init2(mpfb1, precu*53);
  mpfr_init2(mpfb2, precu*53);
  mpfr_init2(mpfb3, precu*53);
  mpfr_init2(mpfb4, precu*53);
	mpfr_set_d(mpfb1, 1.0, GMP_RNDN);
  mpfr_set_d(mpfb2, 1.0, GMP_RNDN);
  mpfr_set_d(mpfb3, 1.0, GMP_RNDN);
  mpfr_set_d(mpfb4, 1.0, GMP_RNDN);

  // Capture the start time.
  timeval tim;
  double t_start, t_stop;
  my_timer(tim, t_start, t_stop, "start");
       
	for (int i=0; i<trials; i++){
		mpfr_add(mpfb1, mpfb1, mpfa1, GMP_RNDN);
		mpfr_add(mpfb2, mpfb2, mpfa2, GMP_RNDN);
		mpfr_add(mpfb3, mpfb3, mpfa3, GMP_RNDN);
		mpfr_add(mpfb4, mpfb4, mpfa4, GMP_RNDN);
	}
	double seconds = my_timer(tim, t_start, t_stop, "stop");
  cout << " Performance: " << 4*1e-6*trials/seconds << " Mops/second." << endl;

  //end timing      
	mpfr_add(mpfb1, mpfb1, mpfb2, GMP_RNDN);
	mpfr_add(mpfb1, mpfb1, mpfb3, GMP_RNDN);
	mpfr_add(mpfb1, mpfb1, mpfb4, GMP_RNDN); 
	buff[0] = mpfr_get_d(mpfb1, GMP_RNDN);

	mpfr_clear(mpfa1);
  mpfr_clear(mpfa2);
  mpfr_clear(mpfa3);
  mpfr_clear(mpfa4);
  mpfr_clear(mpfb1);
  mpfr_clear(mpfb2);
  mpfr_clear(mpfb3);
  mpfr_clear(mpfb4);

  return buff[0];
}

int main(int argc, char *argv[]) {
	if(argc<2){
		cout << "Usage: ./timingAdd_c trials" << endl; 
   	return 0;
 	}

	int trials = atoi(argv[1]);

  cout << "Testing " << trials << " additions with " << precu << "d = " << precu << "d + " << prec << "d precision." << endl;
  double res = testAdd(trials);
  cout << "Result: " << res << endl;
  return 0;
}
