/*
 * timingAdd_DD.cu - Bailey style testing
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2013 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu popescu@laas.fr
 *               Mioara Joldes joldes@laas.fr
 */

#include <ctime>
#include <cstdio>
#include <gmp.h>
#include <iostream>
#include <iomanip>
#include <string>
using namespace std;
#include <gmp.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpfr.h>
#include <mpfi.h>
#include "mpfi_io.h"
#include <qd/qd_real.h>
#include "../../src/multi_prec.h"
#define prec 2

const double MICRO(1e-6);       // Used for the timing.
//----------------------------------------------------------------------
// A basic timer that works for OMP code.
//----------------------------------------------------------------------
double my_timer(timeval &tim, double &t_start, double &t_stop,
                          const string &command){
	if ( 0 == command.compare("start") ) {
  	gettimeofday(&tim, NULL);
    t_start = tim.tv_sec+(tim.tv_usec*MICRO);
  }
  if ( 0 == command.compare("stop") ) {
    gettimeofday(&tim, NULL);
    t_stop = tim.tv_sec+(tim.tv_usec*MICRO);
    cout << "Wall time: " << setprecision(8) << fixed
         << t_stop - t_start << " seconds.";
  }
  return (t_stop - t_start);
}

double testAdd(int trials, int mpf_prec){
	dd_real  mpa, mpa1, mpa2, mpa3, mpa4;
	dd_real b1, b2, b3, b4;
	double buff[prec];
  generateNormalizedExpansion<prec>(buff);
  mpa=(dd_real)(buff[0])+(dd_real)(buff[1]);
	mpa1 = 1.0 + (1.0/trials) * mpa;
	mpa2 = 1.0 + (2.0/trials) * mpa;
	mpa3 = 1.0 + (3.0/trials) * mpa;
  mpa4 = 1.0 + (4.0/trials) * mpa;	
	b1 = 1.0;
  b2 = 1.0;
	b3 = 1.0;
	b4 = 1.0;

  // Capture the start time.
  timeval tim;
  double t_start, t_stop;
  my_timer(tim, t_start, t_stop, "start");
       
	for (int i=0; i<trials; i++){
		b1 += mpa1;
		b2 += mpa2;
		b3 += mpa3;
		b4 += mpa4;
	}
	double seconds = my_timer(tim, t_start, t_stop, "stop");
  cout << " Performance: " << 4*1e-6*trials/seconds << " Mops/second." << endl;

  //end timing      
	b1=(b1+b2+b3+b4);
  return to_double(b1);
}

int main(int argc, char *argv[]) {
	if(argc<2){
		cout << "Usage: ./timingAdd_DD trials" << endl; 
   	return 0;
 	}

	int trials = atoi(argv[1]);

	int mpf_prec = 53*prec*prec;
  
  cout << "Testing QD " << trials << " addition with double-double precision." << endl;
  double res = testAdd(trials, mpf_prec);
  cout << "Result: " << res << endl;
  return 0;
}
