/*
 * multi_prec.h
 * This file is part of MultiPrecGPU Library
 *
 * Copyright (C) 2014 - 
 *
 * MultiPrecGPU Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MultiPrecGPU Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MultiPrecGPU Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */
/* Contributors: Valentina Popescu valentina.popescu@ens-lyon.fr
 *               Mioara Joldes joldes@laas.fr
 */
/* Multi prec class
 */

#ifndef _multi_prec_h
#define _multi_prec_h

#include "Kprecision_op.h"
/**forward declarations**/
template <int prec, class T> class multi_prec ;

/**template friends**/
//********************* ADD *********************//
template <int prec, class T>
__host__ __device__ multi_prec<prec,T> abs(const multi_prec<prec,T> &mp);

template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void addExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);

template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void addExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void truncatedAddExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);

template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void truncatedAddExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

//********************* MUL *********************//
template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void mulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);

template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void mulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void truncatedMulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);

template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void truncatedMulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

//********************* DIV *********************//
template <int pR, class TR, int p1, class T1>
__host__ __device__ void invExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);

template <int pR, class TR, int p1, class T1>
__host__ __device__ void invExpansion(multi_prec<pR,TR> &res, const T1 val);

template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void divExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);

template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void divExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

//********************* SQRT *********************//
template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);

template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtExpansion(multi_prec<pR,TR> &res, const T1 val);

template <int pR, class TR, int p1, class T1>
__host__ __device__ void invSqrtExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);

template <int pR, class TR, int p1, class T1>
__host__ __device__ void invSqrtExpansion(multi_prec<pR,TR> &res, const T1 val);

template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtHeronExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);

template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtHeronExpansion(multi_prec<pR,TR> &res, const T1 val);

template <int prec, class T>
class multi_prec {
private:
	T data[prec];
public:
	//---------constructors--------------
	__host__ __device__ multi_prec(){};
	__host__ __device__ multi_prec(const T x);
	__host__ __device__ multi_prec(const T *datap, const int precp);

	//----------geters & setters--------------
	__host__ __device__ int getPrec() const;
	__host__ __device__ const T* getData() const;
	__host__ __device__ void setData(const T *datap, const int precp);

 	/**pretty print**/
 	__host__ __device__ void printMultiPrec(){
  	printf("Prec = %d\n", prec); 
  	for(int i=0;i<prec;i++) printf("   Data[%d] = %e\n",i,data[i]);
 	}
 	__host__ __device__ void printMultiPrecBin(){
  	printf("Prec = %d\n", prec); 
  	for(int i=0;i<prec;i++)	printf("   Data[%d] = %a;\n",i,data[i]);
 	}
  __host__ __device__ void printMultiPrecBinUnevalSum(){
  	for(int i=0;i<prec-1;i++) printf("%a + ", data[i]);
    printf("%a;", data[prec-1]);
  }

	//----------operators-----------------------
	/** operator [] overloading  **/
	__host__ __device__ const T operator [](const int i)const {return data[i];}

	/** Puts the value other in data[0] of the current multi_prec variable  **/
	__host__ __device__ multi_prec& operator =(const T other);

	/** Transfers the multi_prec parameter other in the current multi_prec variable **/
	__host__ __device__ multi_prec<prec,T>& operator =(const multi_prec<prec,T>& other){
		//check against self assingment
		if(this != &other) setData(other.getData(),prec);   
		return *this;
	}

	/** Transfers the multi_prec parameter other in the current multi_prec variable **/
	template <int pS>
	__host__ __device__ multi_prec& operator =(const multi_prec<pS,T>& other){
	 	setData(other.getData(),pS);
	 	return *this;
	}

	/** Equality overloading**/
	template <int pS>
	__host__ __device__ bool operator ==(const multi_prec<pS,T> &mp2);
	template <int pS>
	__host__ __device__ bool operator !=(const multi_prec<pS,T> &mp2);

	/**Random expansion generation **/
	__host__ __device__ void randomInit(){ generateNormalizedExpansion<prec, T>(data);}

	/**Unary - overloading**/
	__host__ __device__ friend multi_prec operator -(const multi_prec<prec,T> &mp1 ){
		multi_prec<prec,T> res;
		res = multi_prec<prec,T>(0);
		for(int i=0; i<prec; i++) res.data[i] = -mp1.getData()[i];
		return res;
	}

	/** operator += overloading  **/
	template <int pS>
	__host__ 	__device__ multi_prec<prec,T>& operator +=(const multi_prec<pS,T> &mp){
		truncatedAddExpansion<prec,T,prec,T,pS,T>(*this,*this,mp);
		return *this;
	}
	__host__ 	__device__ multi_prec<prec,T>& operator +=(const T val){
		truncatedAddExpansion<prec,T,prec,T,T>(*this,*this,val);
		return *this;
	}

  /** operator -= overloading  **/
  template <int pS>
  __host__ __device__ multi_prec<prec,T>& operator -=(const multi_prec<pS,T> &mp){
    truncatedAddExpansion<prec,T,prec,T,pS,T>(*this,*this,-mp);
    return *this;
  }
  __host__  __device__ multi_prec<prec,T>& operator -=(const T val){
    truncatedAddExpansion<prec,T,prec,T,T>(*this,*this,-val);
    return *this;
  }

  /** operator *= overloading  **/
  template <int pS>
  __host__ __device__ multi_prec<prec,T> operator *=(const multi_prec<pS,T> &mp){
    truncatedMulExpansion<prec,T,prec,T,pS,T>(*this, *this, mp);
    return *this;
  }
  __host__  __device__ multi_prec<prec,T>& operator *=(const T val){
    truncatedMulExpansion<prec,T,prec,T,T>(*this, *this, val);
    return *this;
  }

	/** operator /= overloading  **/
	template <int pS>
	__host__ 	__device__ multi_prec<prec,T>& operator /=(const multi_prec<pS,T> &mp){
		divExpansion<prec,T,prec,T,pS,T>(*this,*this,mp);
		return *this;
	}
	__host__ 	__device__ multi_prec<prec,T>& operator /=(const T val){
		divExpansion<prec,T,prec,T,T>(*this,*this,val);
		return *this;
	}

	/**absolute value **/
	__host__ __device__  friend multi_prec<prec,T> abs<>(const multi_prec<prec,T> &mp);

	/** operator + overloading  **/
	template <int pS>
	__host__ __device__ friend multi_prec operator +(const multi_prec<prec,T> &mp1, const multi_prec<pS,T> &mp2){
		multi_prec<(prec>pS)?prec:pS, T> res(mp1.getData(), prec);
		return res += mp2;
	}
	__host__ __device__ friend multi_prec operator +(const multi_prec<prec,T> &mp1, const T val){
		multi_prec<prec, T> res(mp1.getData(), prec);
		return res += val;
	}
	__host__ __device__ friend multi_prec operator +(const T val, const multi_prec<prec,T> &mp1 ){
		multi_prec<prec, T> res(mp1.getData(), prec);
		return res += val;
	}

	/** operator - overloading  **/
	template <int pS>
	__host__ __device__ friend multi_prec operator -(const multi_prec<prec,T> &mp1, const multi_prec<pS,T> &mp2){
		multi_prec<(prec>pS)?prec:pS, T> res(mp1.getData(), prec);
		return res -= mp2;
	}
	__host__ __device__ friend multi_prec operator -(const multi_prec<prec,T> &mp1, const T val){
		multi_prec<prec, T> res(mp1.getData(), prec);
		return res -= val;
	}
	__host__ __device__ friend multi_prec operator -(const T val, const multi_prec<prec,T> &mp1 ){
		multi_prec<prec, T> res(val);
		return res -= mp1;
	}

	/** operator * overloading  **/
	template <int pS>
	__host__ __device__ friend multi_prec operator *(const multi_prec<prec,T> &mp1, const multi_prec<pS,T> &mp2){
		multi_prec<(prec>pS)?prec:pS, T> res(mp1.getData(), prec);
		return res *= mp2;
	}
	__host__ __device__ friend multi_prec operator *(const multi_prec<prec,T> &mp1, const T val){
		multi_prec<prec, T> res(mp1.getData(), prec);
		return res *= val;
	}
	__host__ __device__ friend multi_prec operator *(const T val, const multi_prec<prec,T> &mp1 ){
		multi_prec<prec, T> res(mp1.getData(), prec);
		return res *= val;
	}

	/** operator / overloading  **/
  template <int pS>
  __host__ __device__ friend multi_prec operator /(const multi_prec<prec,T> &mp1, const multi_prec<pS,T> &mp2){
    multi_prec<(prec>pS)?prec:pS, T> res(mp1.getData(), prec);
    return res /= mp2;
  }
  __host__ __device__ friend multi_prec operator /(const multi_prec<prec,T> &mp1, const T val){
    multi_prec<prec, T> res(mp1.getData(), prec);
    return res /= val;
  }
  __host__ __device__ friend multi_prec operator /(const T val, const multi_prec<prec,T> &mp1 ){
    multi_prec<prec, T> res(mp1.getData(), prec);
    return res /= val;
  }

	/** operator / overloading  **/

	//********************* ADD *********************//
	template <int pR, class TR, int p1, class T1, int p2, class T2>
	__host__ __device__ friend void addExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);
	template <int pR, class TR, int p1, class T1, class T2>
	__host__ __device__ friend void addExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);
	template <int pR, class TR, int p1, class T1, int p2, class T2>
  __host__ __device__ friend void truncatedAddExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);
  template <int pR, class TR, int p1, class T1, class T2>
  __host__ __device__ friend void truncatedAddExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

	//********************* MUL *********************//
	template <int pR, class TR, int p1, class T1, int p2, class T2>
	__host__ __device__ friend void mulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);
	template <int pR, class TR, int p1, class T1, class T2>
	__host__ __device__ friend void mulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);
  template <int pR, class TR, int p1, class T1, int p2, class T2>
  __host__ __device__ friend void truncatedMulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);
  template <int pR, class TR, int p1, class T1, class T2>
  __host__ __device__ friend void truncatedMulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

	//********************* DIV *********************//
	template <int pR, class TR, int p1, class T1>
	__host__ __device__ friend void invExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);
	template <int pR, class TR, int p1, class T1>
	__host__ __device__ friend void invExpansion(multi_prec<pR,TR> &res, const T1 val);
	template <int pR, class TR, int p1, class T1, int p2, class T2>
	__host__ __device__ friend void divExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2);
	template <int pR, class TR, int p1, class T1, class T2>
	__host__ __device__ friend void divExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val);

	//********************* SQRT *********************//
	template <int pR, class TR, int p1, class T1>
	__host__ __device__ friend void sqrtExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);
	template <int pR, class TR, int p1, class T1>
  __host__ __device__ friend void sqrtExpansion(multi_prec<pR,TR> &res, const T1 val);
	template <int pR, class TR, int p1, class T1>
	__host__ __device__ friend void invSqrtExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);
	template <int pR, class TR, int p1, class T1>
  __host__ __device__ friend void invSqrtExpansion(multi_prec<pR,TR> &res, const T1 val);
	template <int pR, class TR, int p1, class T1>
  __host__ __device__ friend void sqrtHeronExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1);
	template <int pR, class TR, int p1, class T1>
  __host__ __device__ friend void sqrtHeronExpansion(multi_prec<pR,TR> &res, const T1 val);
};

/*definitions*/

/** A constructor that initializes the array data with the value x 
    and the remaining positions with 0 **/
template <int prec, class T>
__host__ __device__ multi_prec<prec, T>::multi_prec(const T x){
	data[0] = (T)x;
  for(int i=1; i<prec; i++) data[i] = 0;
}

/** A constructor that initializes the array data with precp values in the array datap
    If precp> prec, the remaining elements of datap are ignored **/
template <int prec, class T>
__host__ __device__ multi_prec<prec, T>::multi_prec(const T *datap,const int precp){
  for(int i=0; i<min(prec, precp); i++)	data[i]=datap[i];
  for(int i=min(prec,precp); i<prec; i++) data[i]=0;
}

//----------geters & setters--------------
template <int prec, class T>
__host__ __device__ int multi_prec<prec, T>::getPrec() const{ return prec; }
template <int prec, class T>
__host__ __device__ const T* multi_prec<prec, T>::getData()const{	return data;}
template <int prec, class T>
__host__ __device__ void multi_prec<prec, T>::setData(const T *datap, const int precp){  
  for(int i=0; i<min(prec, precp); i++)	data[i]=datap[i];
  for(int i=min(prec, precp); i<prec; i++) data[i]=0;
}

//********************* ADD *********************//
/** Adds the two parameters mp1 and mp2 and puts the result in res
    The sum is computed using merge algorithm followed by the renormalization algorithm **/
template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void addExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2){
  certifiedAdd<p1,T1,p2,T2,pR,TR>(mp1.data, mp2.data, res.data);
}
template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void addExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val){
  certifiedAdd<p1,T1,1,T2,pR,TR>(mp1.data, &val, res.data);
}

/** Adds the significant componets of the two parameters mp1 and mp2 and puts the result in res
    The sum is computed using merge algorithm followed by the renormalization algorithm 
		Error bound: |x+y-r| <= (|x0|+|y0|)2^{-(p-1)R} / (1-2^{-(p-1)}) **/
template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void truncatedAddExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2){
  truncatedAdd<p1,T1,p2,T2,pR,TR>(mp1.data, mp2.data, res.data);
}
template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void truncatedAddExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val){
  truncatedAdd<p1,T1,1,T2,pR,TR>(mp1.data, &val, res.data);
}

//********************* MUL *********************//
/** Multiplies the two parameters mp1 and mp2 and puts the result in res
    The result is computed using an adaption of Priest's multiplication algorithm **/
template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void mulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2){
	certifiedMul<p1,T1,p2,T2,pR,TR>(mp1.data, mp2.data, res.data);
}
template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void mulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val){
	certifiedMul<p1,T1,1,T2,pR,TR>(mp1.data, &val, res.data);
}

/** Multiplies the significant components of the two parameters mp1 and mp2 and puts the result in res
		The result is computed using a generalization of Bailey's multiplication algorithm
    Error bound: |xy-r| <=  **/
template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void truncatedMulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2){
  truncatedMul<p1,T1,p2,T2,pR,TR>(mp1.data, mp2.data, res.data);
}
template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void truncatedMulExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val){
  truncatedMul<p1,T1,1,T2,pR,TR>(mp1.data, &val, res.data);
}

//********************* DIV *********************//
/** Computes the reciprocal of parameter mp1 and puts the result in res
    The method used is based on an adapted Newton-Raphson iteration 
		Initial guess q_0 and the iteration q_{n+1} = q_n(2-yq_n) 
		Error bound: |(r-a^{-1})/a^{-1}| <= 2^{-2^q(p-3)-1} **/
template <int pR, class TR, int p1, class T1>
__host__ __device__ void invExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1){
	invNewtonUp<p1,T1,pR,TR>(mp1.data, res.data);
}
template <int pR, class TR, int p1, class T1>
__host__ __device__ void invExpansion(multi_prec<pR,TR> &res, const T1 val){
	invNewtonUp<1,T1,pR,TR>(&val, res.data);
}

/** Divides the two parameters mp1 and mp2 and puts the result in res
    The method used is based on an adapted Newton-Raphson iteration followed by a multiplication
    Initial guess q_0 and the iteration q_{n+1} = q_n(2-yq_n) **/
template <int pR, class TR, int p1, class T1, int p2, class T2>
__host__ __device__ void divExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const multi_prec<p2,T2> &mp2){
  divNewton<p1,T1,p2,T2,pR,TR>(mp1.data, mp2.data, res.data);
}
template <int pR, class TR, int p1, class T1, class T2>
__host__ __device__ void divExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1, const T2 val){
  divNewton<p1,T1,1,T2,pR,TR>(mp1.data, &val, res.data);
}

//********************* SQRT *********************//
/** Computes the square root of parameter mp1 and puts the result in res
    The method used is based on the adapted 'division-free' Newton-Raphson iteration followed by a multiplication
    Initial guess q_0 and the iteration q_{n+1} = (1/2)q_n(3-y(q_n)^2) **/
template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1){
  sqrtDirectNewton<p1,T1,pR,TR>(mp1.data, res.data);
}
template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtExpansion(multi_prec<pR,TR> &res, const T1 val){
  sqrtDirectNewton<1,T1,pR,TR>(&val, res.data);
}

/** Computes the square root of the parameter mp1 and puts the result in res
    The method used is based on the adapted 'division-free' Newton-Raphson iteration
    Initial guess q_0 and the iteration q_{n+1} = (1/2)q_n(3-y(q_n)^2) **/
template <int pR, class TR, int p1, class T1>
__host__ __device__ void invSqrtExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1){
  sqrtInvNewton<p1,T1,pR,TR>(mp1.data, res.data);
}
template <int pR, class TR, int p1, class T1>
__host__ __device__ void invSqrtExpansion(multi_prec<pR,TR> &res, const T1 val){
  sqrtInvNewton<1,T1,pR,TR>(&val, res.data);
}

/** Computes the square root of the parameter mp1 and puts the result in res
    The method used is based on the adapted 'Heron' Newton-Raphson iteration
    Initial guess q_0 and the iteration q_{n+1} = (1/2)(q_n+y/q_n) **/
template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtHeronExpansion(multi_prec<pR,TR> &res, const multi_prec<p1,T1> &mp1){
  sqrtHeron<p1,T1,pR,TR>(mp1.data, res.data);
}
template <int pR, class TR, int p1, class T1>
__host__ __device__ void sqrtHeronExpansion(multi_prec<pR,TR> &res, const T1 val){
  sqrtHeron<1,T1,pR,TR>(&val, res.data);
}

/**-----------operators-------------**/
/** Puts the value other in data[0] of the current multi_prec variable **/
template <int p, class T>
__host__ __device__ multi_prec<p,T>& multi_prec<p,T>::operator = (const T other){
  data[0]=other;	
  for(int i=1; i<p; i++) data[i] = 0;
	return *this;
}

/** Checks equality between to multiprecs.
    If precs are different, the higher precision terms have to be zero **/
template <int p, class T>
template<int pS>
__host__ __device__ bool multi_prec<p,T>::operator ==(const multi_prec<pS,T> &mp2){
	for(int i=0; i<min(p,pS); i++)
	if(data[i] != (mp2.getData())[i]) return false;
  if(p<pS) 
  	for(int i=p;i<pS;i++) if(0 != (mp2.getData())[i])	return false;  
  if(pS<p) 
   	for(int i=pS;i<p;i++) if(0 != data[i]) return false;  
	return true;
}

/** Checks inequality between to multiprecs.
    Works as NOT (operator ==) defined above **/
template <int p, class T>
template <int pS>
__host__ __device__ bool multi_prec<p,T>::operator !=(const multi_prec<pS,T> &mp2){ return !((*this) == mp2); }

/** Returns the absolute value of the multi_prec parameter mp
    Remark: if data[0] < 0 returns the negation of mp **/
template <int p, class T >
__host__ __device__ multi_prec<p,T> abs(const multi_prec<p,T> &mp){
 	multi_prec<p, T> res;
 	res = mp;
	if((res.getData())[0] < 0) return -res;
	return res;
}

#endif
